@@ -13,8 +13,8 @@ import types
 from xml.etree.ElementTree import iterparse
 from datetime import date
 
-from medic.orm import \
-        Medline, Section, Author, Descriptor, Qualifier, Database, Identifier, Chemical
+from medic.orm import Medline, Section, Author, Descriptor, Qualifier, Database, Identifier, \
+        Chemical, Keyword, PublicationType
 
 __all__ = ['MedlineXMLParser', 'PubMedXMLParser']
 
@@ -88,8 +88,7 @@ class Parser:
             version = element.get('VersionID')
 
             if version is not None and version.strip() != "1":
-                logger.info('detected a citation with VersionID "%s"',
-                            version)
+                logger.debug('detected a citation with VersionID "%s"', version)
                 self.skipping()
 
     def yieldInstances(self, element):
@@ -102,7 +101,7 @@ class Parser:
             if element.tag == 'MedlineCitation':
                 self.undefined()
         elif hasattr(self, element.tag):
-            logger.debug('processing %s: %s', element.tag, repr(element.text))
+            logger.debug('processing %s', element.tag)
             instance = getattr(self, element.tag)(element)
 
             if instance is not None:
@@ -110,19 +109,19 @@ class Parser:
 
                 if isinstance(instance, types.GeneratorType):
                     for i in instance:
-                        yield i
+                        if i is not None:
+                            yield i
                 else:
                     yield instance
             else:
                 logger.debug('ignored %s', element.tag)
 
-
     def DeleteCitation(self, element):
         for pmid in element.findall('PMID'):
             yield int(pmid.text)
 
     def MedlineCitation(self, element):
-        dates = {}
+        options = {}
 
         for name, key in (('DateCompleted', 'completed'),
                           ('DateCreated', 'created'),
@@ -130,11 +129,62 @@ class Parser:
             e = element.find(name)
 
             if e is not None:
-                dates[key] = ParseDate(e)
+                options[key] = ParseDate(e)
 
+        created = options['created']
+        del options['created']
         status = element.get('Status')
         journal = element.find('MedlineJournalInfo').find('MedlineTA').text.strip()
-        return Medline(self.pmid, status, journal, **dates)
+        article = element.find('Article')
+        pub_date = self.parsePubDate(article.find('Journal/JournalIssue/PubDate'))
+        issue = self.parseIssue(article.find('Journal/JournalIssue'))
+        pagination = self.parsePagination(article.find('Pagination/MedlinePgn'))
+
+        if issue:
+            options['issue'] = issue
+        if pagination:
+            options['pagination'] = pagination
+
+        return Medline(self.pmid, status, journal, pub_date, created, **options)
+
+    def parsePubDate(self, element):
+        medline = element.find('MedlineDate')
+
+        if medline is not None:
+            return medline.text.strip()
+        else:
+            date = [element.find('Year').text.strip()]
+
+            if element.find('Season') is not None:
+                date.append(element.find('Season').text.strip())
+            elif element.find('Month') is not None:
+                date.append(element.find('Month').text.strip())
+
+                if element.find('Day') is not None:
+                    date.append(element.find('Day').text.strip())
+
+            return ' '.join(date)
+
+    def parseIssue(self, element):
+        if element is not None:
+            issue = None
+            vol = element.find('Volume')
+            iss = element.find('Issue')
+
+            if vol is not None and vol.text:
+                issue = vol.text.strip()
+
+            if iss is not None and iss.text:
+                if issue is None:
+                    issue = iss.text.strip()
+                else:
+                    issue = '{}({})'.format(issue, iss.text.strip())
+
+            return issue
+
+    def parsePagination(self, element):
+        if element is not None and element.text:
+            return element.text.strip()
 
     def PMID(self, element):
         pmid = int(element.text)
@@ -157,44 +207,45 @@ class MedlineXMLParser(Parser):
     def __init__(self, *args, **kwargs):
         super(MedlineXMLParser, self).__init__(*args, **kwargs)
         self.seq = 0
-        self.num = 0
-        self.sub = 0
-        self.pos = 0
-        self.chem = 0
         self.namespaces = None
 
     def reset(self, pmid):
         super(MedlineXMLParser, self).reset(pmid)
         self.seq = 0
-        self.num = 0
-        self.sub = 0
-        self.pos = 0
-        self.chem = 0
         self.namespaces = set()
 
-    def AbstractText(self, element):
-        # sometimes there are non-content AbstractText
-        # elements in MEDLINE/PubMed ("<AbstractText/>")
-        if element.text is not None:
-            text = element.text.strip()
-            # and, less frequently, they might only contain whitespaces
-            if text:
-                self.seq += 1
-                section = element.get('NlmCategory', 'Abstract').capitalize()
-                return Section(
-                    self.pmid, self.seq, section, text, element.get('Label', None)
-                )
+    def Abstract(self, element):
+        for e in element.getchildren():
+            if element.text is not None and element.text.strip():
+                if e.tag == 'CopyrightInformation':
+                    yield self.parseCopyrightInformation(e)
+                else:
+                    yield self.parseAbstractText(e)
 
-        logger.info('empty %s AbstractText in %i',
-                    element.get('NlmCategory', 'ABSTRACT'), self.pmid)
-        return None
+    def parseAbstractText(self, element, other=None):
+        text = element.text.strip()
+        self.seq += 1
+        name = element.get('NlmCategory', 'Abstract').capitalize()
+        name = '{}{}'.format(other, name) if other else name
+        return Section(
+            self.pmid, self.seq, name, text, element.get('Label', None)
+        )
+
+    def parseCopyrightInformation(self, element, other=None):
+        name = '{}Copyright'.format(other) if other else 'Copyright'
+        self.seq += 1
+        return Section(self.pmid, self.seq, name, element.text.strip())
 
     def ArticleTitle(self, element):
         if element.text is not None:
             self.seq += 1
             return Section(self.pmid, self.seq, 'Title', element.text.strip())
 
-    def Author(self, element):
+    def AuthorList(self, element):
+        for pos, author in enumerate(element.getchildren()):
+            yield self.parseAuthor(pos, author)
+
+    def parseAuthor(self, pos, element):
         name, forename, initials, suffix = self.parseAuthorElements(element.getchildren())
 
         if initials == forename and initials is not None:
@@ -202,8 +253,7 @@ class MedlineXMLParser(Parser):
             forename = None
 
         if name is not None:
-            self.pos += 1
-            return Author(self.pmid, self.pos, name, initials, forename, suffix)
+            return Author(self.pmid, pos + 1, name, initials, forename, suffix)
         else:
             logger.warning('empty or missing Author/LastName or CollectiveName in %i',
                            self.pmid)
@@ -218,11 +268,11 @@ class MedlineXMLParser(Parser):
                 if child.tag == 'LastName':
                     name = text
                 elif child.tag == 'ForeName':
-                    forename = text[:128]
+                    forename = text[:256]  # safeguard for size violations
                 elif child.tag == 'Initials':
-                    initials = text[:128]
+                    initials = text
                 elif child.tag == 'Suffix':
-                    suffix = text[:128]
+                    suffix = text
                 elif child.tag == 'CollectiveName':
                     name = text
                     forename = ''
@@ -238,23 +288,22 @@ class MedlineXMLParser(Parser):
             else:
                 logger.warning('empty Author element %s in %i"', child.tag, self.pmid)
 
+        if forename and initials:
+            if forename.replace(' ', '') == initials:
+                forename = None
+
         return name, forename, initials, suffix
 
-    def Chemical(self, element):
-        self.chem += 1
-        e = element.find('RegistryNumber')
-        uid = None
+    def ChemicalList(self, element):
+        for idx, chemical in enumerate(element.getchildren()):
+            e = chemical.find('RegistryNumber')
+            uid = None
 
-        if e is not None and e.text is not None and e.text.strip() != "0":
-            uid = e.text.strip()
+            if e is not None and e.text is not None and e.text.strip() != "0":
+                uid = e.text.strip()
 
-        name = element.find('NameOfSubstance')
-        return Chemical(self.pmid, self.chem, uid, name.text.strip())
-
-    def CopyrightInformation(self, element):
-        if element.text is not None:
-            self.seq += 1
-            return Section(self.pmid, self.seq, 'Copyright', element.text.strip())
+            name = chemical.find('NameOfSubstance')
+            yield Chemical(self.pmid, idx + 1, name.text.strip(), uid)
 
     def DataBank(self, element):
         name = element.find('DataBankName')
@@ -267,23 +316,6 @@ class MedlineXMLParser(Parser):
                     done.add(acc.text)
                     yield Database(self.pmid, name.text, acc.text)
 
-    def DescriptorName(self, element):
-        if element.text is not None:
-            self.num += 1
-            self.sub = 0
-            return Descriptor(
-                self.pmid, self.num, element.text.strip(),
-                (element.get('MajorTopicYN', 'N') == 'Y')
-            )
-
-    def QualifierName(self, element):
-        if element.text is not None:
-            self.sub += 1
-            return Qualifier(
-                self.pmid, self.num, self.sub, element.text.strip(),
-                (element.get('MajorTopicYN', 'N') == 'Y')
-            )
-
     def ELocationID(self, element):
         ns = element.get('EIdType').strip().lower()
 
@@ -291,7 +323,16 @@ class MedlineXMLParser(Parser):
             self.namespaces.add(ns)
             return Identifier(self.pmid, ns, element.text.strip())
 
-        return None
+    def KeywordList(self, element):
+        owner = element.get('Owner', 'NLM').strip().upper()
+        logger.debug('KeywordList Owner="%s"', owner)
+
+        for cnt, keyword in enumerate(element.getchildren()):
+            if keyword.text is not None:
+                yield Keyword(
+                    self.pmid, owner, cnt + 1, keyword.text.strip(),
+                    keyword.get('MajorTopicYN', 'N') == 'Y',
+                )
 
     def MedlineCitation(self, element):
         instance = Parser.MedlineCitation(self, element)
@@ -299,6 +340,39 @@ class MedlineXMLParser(Parser):
         self.undefined()
         return instance
 
+    def MeshHeadingList(self, element):
+        for num, mesh in enumerate(element.getchildren()):
+            descriptor = mesh.find('DescriptorName')
+
+            if descriptor is not None and descriptor.text:
+                yield self.parseDescriptor(num, descriptor)
+
+            for sub, qualifier in enumerate(mesh.findall('QualifierName')):
+                if qualifier.text:
+                    yield self.parseQualifier(num, sub, qualifier)
+
+    def parseDescriptor(self, num, element):
+        return Descriptor(
+            self.pmid, num + 1, element.text.strip(),
+            element.get('MajorTopicYN', 'N') == 'Y',
+        )
+
+    def parseQualifier(self, num, sub, element):
+        return Qualifier(
+            self.pmid, num + 1, sub + 1, element.text.strip(),
+            element.get('MajorTopicYN', 'N') == 'Y',
+        )
+
+    def OtherAbstract(self, element):
+        other = element.get('Type')
+
+        for e in element.getchildren():
+            if element.text is not None and element.text.strip():
+                if e.tag == 'CopyrightInformation':
+                    yield self.parseCopyrightInformation(e, other)
+                else:
+                    yield self.parseAbstractText(e, other)
+
     def OtherID(self, element):
         if element.get('Source', None) == 'NLM':
             text = element.text.strip()
@@ -308,7 +382,9 @@ class MedlineXMLParser(Parser):
                     self.namespaces.add('pmc')
                     return Identifier(self.pmid, 'pmc', text.split(' ', 1)[0])
 
-        return None
+    def PublicationType(self, element):
+        if element.text:
+            return PublicationType(self.pmid, element.text.strip())
 
     def VernacularTitle(self, element):
         if element.text is not None:
